#include  "active_repository.h"
#include  "monitor_supplyer.h"
#include  "log_supplyer.h"
#include  "client_log_debug_message_supplyer.h"
#include  "client_debug_view.h"
#include  "compat_stringstream.h"

#include  <utility>
#include  <string>
#include  <cmath>
#include  <csignal>
#include  <cstdio>
#include  <cstdlib>
#include  <cctype>
#include  <unistd.h>
#include  <stdexcept>
#include  <sys/types.h>
#include  <sys/stat.h>
#include  <sys/param.h>
#include  <errno.h>

const double	Active_Repository::PENALTY_AREA_LENGTH = 16.5 /* meter */;
const double	Active_Repository::PENALTY_AREA_HALF_WIDTH = 20.16 /* meter */;
const double	Active_Repository::FIELD_HALF_LENGTH = 52.5 /* meter */;

const int	Active_Repository::SIMULATOR_STEP_MSEC = 100 /* msec */;
const long	Active_Repository::AUTO_DROP_WAIT_STEP = 150 /* step */;
const double	Active_Repository::AUTO_DROP_CIRCLE_R = 2.0 /* meter */;
const int	Active_Repository::AUTO_DROP_COUNT = 5 /* sec */;
const long	Active_Repository::AUTO_KICK_OFF_WAIT_STEP = 300 /* step */;
const long	Active_Repository::AUTO_QUIT_WAIT_STEP = 30 /* step */;

Active_Repository::Active_Repository
		   ( const ref_count_ptr<const Abstruct_Updator_Factory> &
							  updator_factory ,
		     const std::string &  host_name ,
		     port_number_t  server_port_number ,
		     port_number_t  debug_port_number ,
		     int  server_protocol_version ,
		     bool  log_file_source ,
		     const std::string &  log_file_path ,
		     bool  receive_debug ,
		     bool  interpret_client_log ,
		     const std::string &  client_log_directory ,
		     const std::string &  client_log_viewer_command ,
		     bool  auto_drop_flag ,
		     bool  auto_kick_off_flag ,
		     bool  auto_quit_flag ,
		     bool  auto_kill_server_flag ,
		     long  time_limit ,
		     pid_t  server_process_id )
	: callback( new Active_Repository::Callback( this ) ) ,
	  self_updator( updator_factory -> product() ) ,
	  updator_factory( updator_factory ) ,
	  param() ,
	  changed_receiver() ,
	  view_supplyer() , server_controll_message_handler() ,
	  log_file_source( log_file_source ) ,
	  debug_connection( debug_port_number ,
			    Debug_Server_Connection
					::DEFAULT_MAX_RESPONCE_TIME ,
			    true ) ,
	  receive_debug_flag( receive_debug ) ,
	  client_log_debug_message_handler() ,
	  interpret_client_log_flag( interpret_client_log ) ,
	  client_log_directory( client_log_directory ) ,
	  client_log_viewer_command( client_log_viewer_command ) ,
	  game_over_flag( false ),
	  monitor_responsive( false ) ,
	  server_time( 0 ) ,
	  game_log() ,
	  goal_time() ,
	  team_name_set() ,
	  old_play_mode( Monitor_View_Data::Monitor_Play_Mode::Null_Mode ) ,
	  old_left_team_score( 0 ) , old_right_team_score( 0 ) ,
	  ball_track( AUTO_DROP_WAIT_STEP ) ,
	  last_kick_player_side( S_Side_LR::Unknown ) ,
	  auto_drop_flag( auto_drop_flag && ! log_file_source ) ,
	  auto_drop_count( 0 ) ,
	  auto_kick_off_flag( auto_kick_off_flag && ! log_file_source ) ,
	  auto_kick_off_count( 0 ) ,
	  auto_quit_flag( auto_quit_flag ) , auto_quit_count( 0 ) ,
	  auto_kill_server_flag( auto_kill_server_flag ) ,
	  time_limit( time_limit ) ,
	  server_process_id( server_process_id )
{
	if ( log_file_source )
	{
		this -> server_controll_message_handler
			= static_cast<Interactive_Monitor_View_Supplyer *>(0);

		this -> view_supplyer
			= new Log_Supplyer( updator_factory -> product() ,
					    callback ,
					    log_file_path );
	}
	else
	{
		this -> server_controll_message_handler
			= new Monitor_Supplyer
				( updator_factory -> product() ,
				  callback ,
				  host_name , server_port_number ,
				  server_protocol_version );

		this -> view_supplyer = server_controll_message_handler;
	}

	self_updator -> set_interval_callback( this -> callback ,
					       1000 /* msec */ );
}

Active_Repository::~Active_Repository()
{
}

void   Active_Repository::add_changed_event_receiver
		( const ref_count_ptr<Changed_Event_Receiver> &  receiver )
{
	this -> changed_receiver.push_back( receiver );
}

void   Active_Repository::changed_notify() const
{
	for ( size_t  i = 0  ;  i < this -> changed_receiver.size()  ;  i ++ )
	{
		this -> changed_receiver[i] -> notify();
	}
}

const std::map< long , ref_count_ptr<View_Snapshot> > &
					Active_Repository::get_game_log() const
{
	return( this -> game_log );
}

long   Active_Repository::get_server_time() const
{
	return( this -> server_time );
}

const std::vector< std::pair<long,S_Side_LR> > &
	Active_Repository::get_goal_time() const
{
	return( this -> goal_time );
}

bool   Active_Repository::game_over() const
{
	return( this -> game_over_flag );
}

bool   Active_Repository::monitor_connection_responsive() const
{
	return( this -> view_supplyer -> responsive() );
}

void   Active_Repository::update()
{
	//
	// receive monitor message
	//
	do
	{
		bool	have_s_param;
		SServer_Monitor_Log_Format::server_param_t	s_param;

		ref_count_ptr<const Monitor_View_Data>
			m = view_supplyer -> get( &have_s_param , &s_param );

		if ( have_s_param )
		{
			param.overwrite_server_param
			  ( s_param.kick_power_rate
			      / SServer_Monitor_Log_Format::SHOWINFO_SCALE_2 ,
			    s_param.stamina_max
			      / SServer_Monitor_Log_Format::SHOWINFO_SCALE_2 ,
			    s_param.stamina_inc_max
			      / SServer_Monitor_Log_Format::SHOWINFO_SCALE_2 );
		}


		if ( ! m )
		{
			break;
		}

		long	t = m -> time_count;

		if ( t < 0 )
		{
			t = 0;
		}

		server_time = t;

		if ( ! game_log[ t ] )
		{
			game_log[ t ] = new View_Snapshot();
		}

		game_log[ t ] -> set_monitor_view_data( m );

		if ( old_play_mode != m -> play_mode
		  && m -> play_mode
		     == Monitor_View_Data::Monitor_Play_Mode::Before_Kick_Off )
		{
			auto_kick_off_count = 0;
		}

		if ( auto_kick_off_flag
		  && m -> play_mode
		     == Monitor_View_Data::Monitor_Play_Mode::Before_Kick_Off
		  && auto_kick_off_count >= 0 )
		{
			if ( (++ auto_kick_off_count)
			     >= AUTO_KICK_OFF_WAIT_STEP )
			{
				this -> send_start();
			}
		}

		if ( old_play_mode != m -> play_mode
		  || m -> play_mode
		     == Monitor_View_Data::Monitor_Play_Mode::Time_Over )
		{
			ball_track.clear();
		}

		if ( auto_drop_flag
		  && m -> play_mode
		     != Monitor_View_Data::Monitor_Play_Mode::Before_Kick_Off )
		{
			ball_track.push_front( m -> ball );

			while( ball_track.size()
			       > static_cast<size_t>(AUTO_DROP_WAIT_STEP) )
			{
				std::list<D2_Vector>::iterator
					it = ball_track.end();
				it --;
				ball_track.erase( it );
			}

			auto_drop_count = 0;
			std::list<D2_Vector>::iterator	it;
			for ( it = ball_track.begin()  ;
			      it != ball_track.end()  ;  it ++ )
			{
				if ( (*it - (m -> ball)).r()
				     >= AUTO_DROP_CIRCLE_R )
				{
					break;
				}

				auto_drop_count ++;
			}

			if ( auto_drop_count == AUTO_DROP_WAIT_STEP )
			{
				this -> auto_drop_ball
					  ( m -> ball ,
					    m -> play_mode );

			}
		}

		//
		// set goal time
		//
		if ( m -> play_mode
		     == Monitor_View_Data::Monitor_Play_Mode::Goal_Left
		  || m -> play_mode
		     == Monitor_View_Data::Monitor_Play_Mode::Goal_Right
		  || (m -> play_mode
		      == Monitor_View_Data::Monitor_Play_Mode::Time_Over
		      && (m -> team[0].score != old_left_team_score
		       || m -> team[1].score != old_right_team_score)) )
		{
			S_Side_LR	side;

			if ( m -> play_mode
			     == Monitor_View_Data::Monitor_Play_Mode
								::Goal_Left
			 || (m -> play_mode
			     == Monitor_View_Data::Monitor_Play_Mode::Time_Over
			     && (m -> team[0].score - old_left_team_score)
			      > (m -> team[1].score - old_right_team_score) ) )
			{
				side = S_Side_LR::Left_Side;
			}

			if ( m -> play_mode
			     == Monitor_View_Data::Monitor_Play_Mode
								::Goal_Right
			 || (m -> play_mode
			     == Monitor_View_Data::Monitor_Play_Mode::Time_Over
			     && (m -> team[0].score - old_left_team_score)
			      < (m -> team[1].score - old_right_team_score) ) )
			{
				side = S_Side_LR::Right_Side;
			}
			else
			{
				side = S_Side_LR::Unknown;
			}

			if ( std::find( goal_time.begin() ,
					goal_time.end() ,
					std::pair<long,S_Side_LR>
						( m -> time_count , side ) )
			     == goal_time.end() )
			{
				goal_time
				.push_back( std::pair<long,S_Side_LR>
						( m -> time_count , side ) );
			}
		}

		old_play_mode = m -> play_mode;
		old_left_team_score  = m -> team[0].score;
		old_right_team_score = m -> team[1].score;


		if ( m -> play_mode
			    == Monitor_View_Data::Monitor_Play_Mode::Time_Over
		  || (time_limit >= 0  &&  m -> time_count >= time_limit) )
		{
			game_over_flag = true;
		}


		//
		// set last kicked side
		//
		{
			bool	left_kicked = false;
			bool	right_kicked = false;

			for ( int  i = 0  ;  i < MAX_PLAYER  ;  i ++ )
			{
				if ( m -> player[i].kicking()
				  && ! m -> player[i].kicking_fault() )
				{
					left_kicked = true;
					break;
				}
			}

			for ( int  i = MAX_PLAYER  ;  i < MAX_PLAYER * 2  ;
			      i ++ )
			{
				if ( m -> player[i].kicking()
				  && ! m -> player[i].kicking_fault() )
				{
					right_kicked = true;
					break;
				}
			}


			if ( m -> play_mode
			     != Monitor_View_Data::Monitor_Play_Mode::Play_On )
			{
				this -> last_kick_player_side
					= S_Side_LR::Unknown;
			}
			else
			{
				if ( left_kicked && right_kicked )
				{
					this -> last_kick_player_side
						= S_Side_LR::Unknown;
				}
				else if ( left_kicked && ! right_kicked )
				{
					this -> last_kick_player_side
						= S_Side_LR::Left_Side;
				}
				else if ( ! left_kicked && right_kicked )
				{
					this -> last_kick_player_side
						= S_Side_LR::Right_Side;
				}
			}
		}


		//
		// auto kill server
		//
		if ( game_over_flag && auto_kill_server_flag )
		{
			static_cast<void>( this -> kill_server() );

			// XXX
			if ( auto_quit_flag )
			{
				std::exit( 0 );
			}
		}


		//
		// auto quit
		//
		if ( auto_quit_flag && game_over_flag )
		{
			if ( (++ auto_quit_count) >= AUTO_QUIT_WAIT_STEP )
			{
				exit( 0 );
			}
		}


		// open client debug log
		for ( int i = 0  ;  i < 2  ;  i ++ )
		{
			const std::string	team_name
						  = m -> team[i].team_name;

			if ( interpret_client_log_flag
			  && log_file_source /* XXX */
			  && team_name.length() != 0
			  && team_name_set.find( team_name )
			     == team_name_set.end() )
			{
				team_name_set.insert( team_name );

				for ( int  num = 1  ;  num <= MAX_PLAYER  ;
				      num ++ )
				{
					std::string	file_path;

					if ( ! (this
						-> get_client_log_file_path
						     ( &file_path ,
						       team_name , num ) ) )
					{
						continue;
					}

					client_log_debug_message_handler
					.push_back
					( new Client_Log_Debug_Message_Supplyer
					      ( updator_factory -> product() ,
						callback ,
						file_path ) );
				}
			}
		}

		this -> changed_notify();

	} while( false );


	bool	resp = this -> monitor_connection_responsive();

	if ( monitor_responsive != resp )
	{
		this -> changed_notify();
	}

	monitor_responsive = resp;



	//
	// receive debug message
	//
	std::string	debug_message;

	while( this -> receive_debug_flag
	    && this -> debug_connection.recv( &debug_message , false ) )
	{
		this -> register_debug_message( debug_message );
	}


	//
	// read client debug log
	//
	if ( this -> interpret_client_log_flag )
	{
		for ( size_t  i = 0  ;
		      i < this -> client_log_debug_message_handler.size()  ;
		      i ++ )
		{
			if ( this -> client_log_debug_message_handler[i]
							     -> responsive() )
			{
				this -> client_log_debug_message_handler[i]
						     -> get( &debug_message );

				int	n = 0;
				char	left_bracket = '\0';

				if ( std::sscanf( debug_message.c_str() ,
						  "%%%% debug %c%n" ,
						  &left_bracket , &n ) != 1 )
				{
					continue;
				}
				else
				{
					if ( left_bracket != '['
					  || debug_message.rbegin()
					              == debug_message.rend()
					  || *(debug_message.rbegin()) != ']' )
					{
						std::cerr
						  << "Bad log format["
						  << debug_message << "]"
						  << std::endl;

						continue;
					}
				}

				std::string	mes;
				mes.assign( debug_message , 0 , n );
				mes.erase( mes.end() - 1 );

				this -> register_debug_message( mes );
			}
		}
	}
}

// internal
bool   Active_Repository::register_debug_message
					( const std::string &  debug_message )
{
	ref_count_ptr<Client_Debug_View>
		c = new Client_Debug_View( &param );

	if ( ! (c -> decode( debug_message )) )
	{
		std::cerr << "Bad user debug message format["
			  << debug_message << "]" << std::endl;

		return( false );
	}

	if ( c -> field_snapshot() )
	{
		long	t = c -> field_snapshot()
				     -> game_state.current_time.main_step();

		if ( ! this -> game_log[ t ] )
		{
			this -> game_log[ t ] = new View_Snapshot();
		}

		this -> game_log[ t ]
			-> set_client_view
				( new Debug_Client_Field_Recog( *c ) );
	}

	return( true );
}


bool   Active_Repository::have_server_process_id() const
{
	return( this -> server_process_id != -1 );
}

bool   Active_Repository::kill_server() const
{
	if ( this -> have_server_process_id() )
	{
		if ( ::kill( server_process_id , SIGTERM ) == -1 )
		{
			std::perror( "kill" );

			return( false );
		}

		return( true );
	}

	return( false );
}

void   Active_Repository::open_client_log( const std::string &  team_name ,
					   int  uniform_number ,
					   int  step_count ) const
{
	if ( this -> client_log_viewer_command.length() != 0 )
	{
		std::string	file_path;

		if ( ! this -> get_client_log_file_path( &file_path ,
							 team_name ,
							 uniform_number ) )
		{
			return;
		}

		switch( ::fork() )
		{
		case -1:
		      {
			std::perror( "fork" );
			break;
		      }

		case 0:
		      {
			compat_ostringstream	step_string;
			step_string << step_count;

			::setenv( "STEP_COUNT" ,
				  step_string.str().c_str() ,
				  1 );

			::execlp( this -> client_log_viewer_command.c_str() ,
				  this -> client_log_viewer_command.c_str() ,
				  file_path.c_str() ,
				  static_cast<char *>( 0 ) );

			std::perror( "exec" );
			::_exit( 1 );

			break;
		      }

		default:
		      {
			break;
		      }
		}
	}
}

S_Side_LR  Active_Repository::last_kick_side() const
{
	return( this -> last_kick_player_side );
}

const SServer_Param &  Active_Repository::sserver_param() const
{
	return( this -> param );
}

int    Active_Repository::send_start()
{
	this -> auto_kick_off_count = -1;

	if ( this -> server_controll_message_handler )
	{
		return( this -> server_controll_message_handler
							 -> send_start() );
	}
	else
	{
		return( -1 );
	}
}

int    Active_Repository::send_foul( double  x ,  double  y ,
				     S_Side_LR  side )
{
	this -> auto_drop_count = 0;

	if ( this -> server_controll_message_handler )
	{
		return( this -> server_controll_message_handler
						-> send_foul( x , y , side ) );
	}
	else
	{
		return( -1 );
	}
}


//
// Auto Drop
//
void   Active_Repository::set_auto_drop( bool  f )
{
	if ( this -> auto_drop_flag != f )
	{
		ball_track.clear();
	}

	this -> auto_drop_flag = f;
	this -> auto_drop_count = 0;
}

bool   Active_Repository::auto_drop_mode() const
{
	return( this -> auto_drop_flag );
}

long   Active_Repository::auto_drop_rest_step() const
{
	return( Active_Repository::AUTO_DROP_WAIT_STEP
			- this -> auto_drop_count );
}

long   Active_Repository::auto_drop_rest_second() const
{
	if ( this -> auto_drop_rest_step() == static_cast<long>(0) )
	{
		return( static_cast<long>(0) );
	}
	else
	{
		return( static_cast<long>
			( (this -> auto_drop_rest_step() - 1)
			   * SIMULATOR_STEP_MSEC / 1000
			  + 1 ) );
	}
}


//
// Auto Kick Off
//
void   Active_Repository::set_auto_kick_off( bool  f )
{
	this -> auto_kick_off_flag = f;
	this -> auto_kick_off_count = 0;
}

bool   Active_Repository::auto_kick_off_mode() const
{
	return( this -> auto_kick_off_flag );
}

long   Active_Repository::auto_kick_off_rest_step() const
{
	if ( this -> auto_kick_off_count < static_cast<long>(0) )
	{
		return( static_cast<long>(0) );
	}
	else
	{
		return( Active_Repository::AUTO_KICK_OFF_WAIT_STEP
			- this -> auto_kick_off_count );
	}
}

long   Active_Repository::auto_kick_off_rest_second() const
{
	if ( this -> auto_kick_off_rest_step() == static_cast<long>(0) )
	{
		return( static_cast<long>(0) );
	}
	else
	{
		return( static_cast<long>
			( (this -> auto_kick_off_rest_step() - 1)
			   * Active_Repository::SIMULATOR_STEP_MSEC / 1000
			  + 1 ) );
	}
}


//
// Auto Quit
//
void   Active_Repository::set_auto_quit( bool  f )
{
	this -> auto_quit_flag = f;
	this -> auto_quit_count = 0;
}

bool   Active_Repository::auto_quit_mode() const
{
	return( this -> auto_quit_flag );
}

long   Active_Repository::auto_quit_rest_step() const
{
	if ( this -> auto_quit_count < static_cast<long>(0) )
	{
		return( static_cast<long>(0) );
	}
	else
	{
		return( Active_Repository::AUTO_QUIT_WAIT_STEP
			- this -> auto_quit_count );
	}
}

long   Active_Repository::auto_quit_rest_second() const
{
	if ( this -> auto_quit_rest_step() == static_cast<long>(0) )
	{
		return( static_cast<long>(0) );
	}
	else
	{
		return( static_cast<long>
			( (this -> auto_quit_rest_step() - 1)
			   * Active_Repository::SIMULATOR_STEP_MSEC / 1000
			  + 1 ) );
	}
}


//
// Auto Kill Server
//
void   Active_Repository::set_auto_kill_server( bool  f )
{
	this -> auto_kill_server_flag = f;
}

bool   Active_Repository::auto_kill_server_mode() const
{
	return( this -> auto_kill_server_flag );
}




int   Active_Repository::auto_drop_ball
		( const D2_Vector &  ball ,
		  const Monitor_View_Data::Monitor_Play_Mode &  play_mode )
{
	ball_track.clear();

	D2_Vector	drop_point = ball;

	if ( std::fabs( ball.y() ) <= PENALTY_AREA_HALF_WIDTH
	  && std::fabs( ball.x() ) >= FIELD_HALF_LENGTH
					     - PENALTY_AREA_LENGTH )
	{
		double	drop_x;
		double	drop_y;

		drop_x = ((ball.x() >= 0) ?
			    + (FIELD_HALF_LENGTH - PENALTY_AREA_LENGTH)
			  : - (FIELD_HALF_LENGTH - PENALTY_AREA_LENGTH));

		drop_y = ((ball.y() >= 0) ?
			    + PENALTY_AREA_HALF_WIDTH
			  : - PENALTY_AREA_HALF_WIDTH );

		drop_point.set( drop_x , drop_y );
	}

	S_Side_LR	which_ball;
	switch( play_mode.advantage() )
	{
	case S_Side_LR::Left_Side:
		which_ball = S_Side_LR::Right_Side;
		break;

	case S_Side_LR::Right_Side:
		which_ball = S_Side_LR::Left_Side;
		break;

	case S_Side_LR::Unknown:
	default:
		which_ball = S_Side_LR::Unknown;
		break;
	}

	return( this -> send_foul( drop_point.x() , drop_point.y() ,
				   which_ball ) );
}


bool   Active_Repository::get_client_log_file_path
			  ( std::string *  path ,
			    const std::string &  team_name ,
			    int  uniform_number ) const
{
	compat_ostringstream	file_path;

	if ( this -> client_log_directory.length() == 0 )
	{
		char	current_working_directory[MAXPATHLEN];

		if ( ::getcwd( current_working_directory ,
			       sizeof( current_working_directory ) ) != NULL )
		{
			file_path << current_working_directory;
		}
		else
		{
			file_path << ".";
		}

		file_path << "/";
	}
	else
	{
		file_path << this -> client_log_directory;

		if ( *(this -> client_log_directory.rbegin()) != '/' )
		{
			file_path << "/";
		}
	}

	file_path << team_name << "-" << uniform_number << ".slog";

	path -> assign( file_path.str() );

	struct stat	stat_buf;
	if ( ::stat( path -> c_str() , &stat_buf ) == -1 )
	{
		if ( errno != ENOENT )
		{
			std::perror( "stat" );
		}

		return( false );
	}

	return( true );
}
